from logging import getLogger, FileHandler, Formatter

from ConfigParser import ConfigParser


class PidFile(File):
  """
  Contains a numeric process id on each line.  Sometimes it is a singleton, with
  only one pid on the first line, and sometimes it has multiple pids, one per
  line.
  """
  def __init__(self, filename, mode="rw"):
    File.__init__(self, filename, mode)

  def getPid(self):
    "This assumes that this file only contains one pid, on the first line."
    self.seek(0)
    line = self.readline.strip()
    return line.isdigit() and int(line) #or None
  
  def getPids(self):
    "This assumes that there are multiple pids, one per line."
    self.seek(0)
    return [ int(line) for line in [ l.strip() for l in self.readlines() ] if line.isdigit() ]


class ConfigFile(File, ConfigParser):
  "Configuration file, using the ConfigParser module"

  # super class methods are searched left to right, when we use .write(),
  # we want File.write(), not ConfigParser.write()

  # Also keep in mind that since we are also an instance of File, and therefore
  # file, self is a file pointer, so it gets used not only as the instance of
  # self, but a file pointer for ConfigParser to write to.

  def __init__(self, filename, mode="rw"):
    File.__init__(self, filename, mode)
    ConfigParser.__init__(self)
    self.readfp(self)

  def save(self):
    "Saves the state of the configuration back to the config file."
    ConfigParser.write(self, self)


class Playlist(File):
  "Contains path to a file on each line"

  def __init__(self, filename, mode="rw"):
    File.__init__(self, filename, mode)

  getSong = lambda self: self.readline().strip()
  getAllSongs = lambda self: [ line.strip() for line in self.readlines() ]


class AccessFile(File):

  def __dict_from_line(self, line):
    items = [ item.strip() for item in line.split(":@@:") ]
    return { "time": items[0], "song": items[1] }

  def logAccess(self, song):
    self.write("%s :@@: %s\n" % (asctime(), song))
  
  def getSong(self):
    return self.__dict_from_line(self.readStrippedLine())

  def getAllSongs(self):
    return [ self.__dict_from_line(line) for line in self.readStrippedLines() ]

class ErrorLog(File):

  def __init__(self, mode="a"):
    File.__init__(self, filenames["errorlog"], mode)
  
  def writeError(self, msg):
    self.write("%s %s (%s): %s\n" % (asctime(), argv[0].split("/")[-1], getpid(), msg))
    self.flush()

  def unknownCommand(self, command):
    self.writeError("Received unknown command `%s'" % command)

  def decoderHung(self):
    self.writeError("Decoder is hung")

  def die(self, msg=None):
    if msg: self.writeError(msg)
    self.writeError("Exiting")

