from os.path import exists, isfile
from threading import Thread, Event, RLock
from ao import AudioDevice, aoError
from mad import MadFile

from Threading import Repeater

class Decoder(object):
  """
  This is our all-purpose decoder.  Pass it the driver name, and you're ready
  to start playing files:

  myD = Decoder("sun")
  myD.play("path/to/an.mp3")

  You will likely want to call play() in a thread, so you can have some control
  over pausing/stopping/seeking etc. while the song is playing.  This should be
  thread-safe. (it is for me .. so far! :)
  
  Functionality is mostly complete:

  action        function
  ------        --------
  play          play()
  pause         pause()
  resume        resume()
  toggle pause  togglePause()
  stop          stop()
  ffwd          seek(positive int)
  rewind        seek(negative int)

  grabDSP() and ungrabDSP() take and release control of the sound device,
  respctively.  This makes it simple to give up the sound device to other
  processes when needed, without having to exit, or delete the decoder.
  """
  
  # "bits"

  stopRequest   = Event()
  pauseRequest  = Event()
  resumeRequest = Event()

  stopped = Event()
  paused  = Event()
  resumed = Event()

  # end "bits"

  def __init__(self, driver, rate=44100, bits=16):
    """
    Decoder(driver[, rate[, bits]]) -> Decoder object

    The only real action taken upon initialization of Decoder is grabbing and
    then immediately ungrabbing the sound device, so as to immediately
    determine any problem with it.  This actually may change, maybe we want
    to be able to create a Decoder while the sound device is being used.
    """

    self.driver = driver
    self.rate   = rate
    self.bits   = bits

    self.dsp = None

    self.stopped.set()
    self.resumed.set()

    self.playLock  = RLock()
    self.stopperLock  = RLock()

    # testing, testing, is this thing on?  hello?
    self.grabDSP()
    self.ungrabDSP()

  #

  def ungrabDSP(self, ignoreStop=False):
    """
    ungrabDSP() -> None.  Frees sound device.

    Unsets self.dsp, freeing the sound device for use by other processes. If
    a song is currently playing, it will be stopped via stop().
    """
    if not ignoreStop:
      if not self.stopped.isSet():
        self.stopperLock.acquire()
        try:
          self.stop()
        finally:
          self.stopperLock.release()

    if self.dsp:
      self.dsp = None


  def grabDSP(self):
    """
    grabDSP() -> None.  Grabs sound device.

    Sets self.dsp to an ao.AudioDevice(), thereby "grabbing" the sound device
    for use.
    """
    self.dsp = AudioDevice(self.driver, bits=self.bits, rate=self.rate)


  #

  def getBuf(self):
    """
    getBuf() -> True or None.  Reads next chunk of self.madfile into self.buf.

    Reads a hunk of the current mp3 into self.buf.  Returns True if there is
    more song to play, or None if the song has reached its end.
    """

    self.buf = self.madfile.read()
    if self.buf:
      return True
  
  def outputFrame(self):
    "Sends current buffer to sound device"
    self.dsp.play(self.buf)

  def seek(self, position):
    "seek(position_in_seconds) -> None.  Goes to specified position in song."

    if self.stopped.isSet():
      return

    if position < 0:
      position = self.getSongLen() - abs(position)

    #try:
    self.madfile.seek_time(position*1000)
    #except:
      #return

  #

  def _clearData(self):
    "_clearData() -> None.  Resets per-song variables madfile, buf, & filename"

    self.madfile = None
    self.buf = None
    self.filename = str()

  #

  getSongLen        = lambda self: self.madfile.total_time()/1000
  getSongBitrate    = lambda self: self.madfile.bitrate()/1000
  getSongSampleRate = lambda self: self.madfile.samplerate()

  #

  def getBit(self, bit):
    return bit.isSet() and True or False

  def setBit(self, bit, state):
    if state is True:
      bit.set()
    elif state is False:
      bit.clear()
    else:
      raise AttributeError, "state needs to be a boolean"

  #

  def _pause(self):
    self.pauseRequest.clear()
    self.paused.set()

    self.resumeRequest.wait()
    self.resumeRequest.clear()

    self.paused.clear()
    self.resumed.set()

  def stop(self):
    # can't stop if stopped already
    if self.stopped.isSet():
      return

    self.stopRequest.set()

    # get out of the wait() in _pause()
    if self.paused.isSet():
      self.resume()

    self.stopped.wait()


  def pause(self):
    # can't pause if stopped or already paused
    if not self.playLock._RLock__count:
      return

    self.pauseRequest.set()
    self.paused.wait()

    self.resumed.clear()

  def resume(self):
    # can't resume if stopped or already resumed
    if not self.paused.isSet():
      return

    self.resumeRequest.set()
    self.resumed.wait()


  def togglePause(self):
    if self.paused.isSet():
      self.resume()
    elif self.playLock._RLock__count:
      self.pause()

  #

  def play(self, filename):
    if not filename or not isinstance(filename, str):
      raise AttributeError, "Need a proper filename, got `%s' instead" % filename
      return
    if not exists(filename):
      raise IOError, "`%s' does not exist" % filename
      return
    if not isfile(filename):
      raise IOError, "`%s' is not a regular file" % filename
      return


    thread = Thread(target=lambda: self._play(filename))
    thread.setDaemon(True)
    thread.start()


  def _play(self, filename):
    """
    play([filename]) -> None.  Play an mp3 file.

    If filename is empty or a non-string, AttributeError will be raised.
    Otherwise, the file will be loaded by mad, and looped through, checking for
    stop requests at every decoded chunk (which occur tens of times per second).

    play() cleans up after itself when done by calling _clearData()
    """

    self.stopperLock.acquire()

    try:
      if not self.stopped.isSet():
        self.stop()

      self.playLock.acquire()

    finally:
      self.stopperLock.release()

    try:
      self.grabDSP()
      self.stopped.clear()

      self.madfile = MadFile(filename)

      while self.getBuf():
        if self.pauseRequest.isSet():
          self._pause()
        if self.stopRequest.isSet():
          self.stopRequest.clear()
          break

        self.outputFrame()

    finally:
      self._clearData()
      self.ungrabDSP(True)
      self.playLock.release()
      self.stopped.set()




class ThumpDecoder(Decoder):

  """
  ThumpDecoder is a slightly modified Decoder; it logs, takes care of the driver
  name for us [ and other undertermined stuff ].
  """

  debug = lambda self, msg: self.log and self.log.debug("ThumpDecoder: %s" %msg)

  def __init__(self, log=None):
    """
    ThumpDecoder([log]) -> ThumpDecoder object

    The optional log parameter should be a logger object from logger.getLogger()
    """

    self.log = log
    self.debug("__init__: running...")

    Decoder.__init__(self, "oss") # read from config

    self.debug("__init__: done")

  def wrap(self, function, name):
    "wrap(function, name) -> None.  Wraps a function with logging."
    self.debug("%s: running..." % name)
    function()
    self.debug("%s: done" % name)
  
